/* cc2y4m.c - main module
 *
 * Copyright (C) 2003 Rik Snel <snel@phys.uu.nl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; only version 2 of the License, not
 * any other version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 * $Id: cc2y4m.c,v 1.4 2003/12/21 11:10:10 rsnel Exp $
 */
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <pthread.h>
#include <unistd.h>
#include <assert.h>
#include "cc2y4m.h"
#include "bttv.h"
#include "lock.h"
#include "esnelmDecode.h"
#include <SDL/SDL.h>

static void yuv4mpeg2_out(unsigned char* yuv, int fields, FILE *fp) {
	int i;
	fprintf(fp, "FRAME\n");
	fwrite(yuv, 1, 384*288*fields, fp);
	for (i = 0; i < 144; i++) {
		fwrite(yuv+ 384*(288+i)*fields, 1, 192, fp);
		if (fields == 2) fwrite(yuv + 384*(576+2*i) + 192, 1, 192, fp);
	}
	for (i = 0; i < 144; i++) {
		fwrite(yuv+ 384*(432+i)*fields, 1, 192, fp);
		if (fields == 2) fwrite(yuv + 384*(864+2*i) + 192, 1, 192, fp);
	}
}

extern char *optarg;
extern int optind, opterr, optopt;

int quiet = 0, debug = 0;

#define POLX		1975
#define POLY		50
#define POLTRESH	16
#define POLNORM		0
static void poldetect(char *in, int offset) {
	if (offset < 544320 && offset > -30) {
		if (*(in + offset + POLY*2268 + POLX) > POLTRESH) {
			setChroma(1-POLNORM);
			setPolarity(1-POLNORM);
		} else {
			setChroma(POLNORM);
			setPolarity(POLNORM);
		}
	} else {
		setChroma(0);
		setPolarity(0);
	}
}

int main(int argc, char **argv) {
	SDL_Color colors[256];
	SDL_Surface *screen;
	char *out;
	int j;
	unsigned char yuv[4*384*288];
	bttv_t bttv;
	lock_t lock;
	pthread_t capture;
	pthread_t lockthread;
	pthread_attr_t attr;
	int i, k, opt, offset = 0, fields = 1;
	char *in, *field, *dev = "/dev/video0";

	while ((opt = getopt(argc, argv, "VhDqd:i")) != -1) {
		if (opt == '?' || opt == ':') exit(1);
		switch (opt) {
			case 'V':
				printf("cc2y4m version is " VERSION ", (C) 2003 Rik Snel\nPAL color decoding (C) 1999 Ewald Snel.\nThis is free software; see the source for copying conditions. The is NO\nwarranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n");
				exit(1);
			case 'h':
				printf("Example: cc2y4m -i | mplayer - -zoom\n");
				exit(1);
			case 'D':
				debug = 1;
				break;
			case 'q':
				quiet = 1;
				break;
			case 'd':
				dev = optarg;
				break;
			case 'i':
				fields = 2;
				break;
		}

	}
	if (optind != argc) {
		ERROR("unrecognized stuff on the command line, please fix\n");
		exit(1);
	}

	if (bttv_init(&bttv, dev) == -1) {
		exit(1);
	}

	lock_init(&lock, &bttv);

	if (debug) {
		if (SDL_Init(SDL_INIT_VIDEO) < 0) {
			ERROR("Unable to init SDL: %s\n", SDL_GetError());
			exit(1);
		}
		atexit(SDL_Quit);

		for (i = 0; i < 256; i++) {
			colors[i].r = i;
			colors[i].g = i;
			colors[i].b = i;
		}

		colors[255].r = 255;
		colors[255].g = 127;
		colors[255].b = 127;
		colors[254].r = 127;
		colors[254].g = 255;
		colors[254].b = 127;

		screen = SDL_SetVideoMode(300, 200, 8, SDL_HWSURFACE);
		if (!screen) {
			ERROR("%s\n", SDL_GetError());
			exit(1);
		}

		if (SDL_MUSTLOCK(screen)) {
			ERROR("SDL screen must be locked, but locking not implemented (lazy)\n");
			exit(1);
		}

		SDL_SetColors(screen, colors, 0, 256);

		out = (Uint8*)screen->pixels;
	}

	InitializeDecoderMMX();

	pthread_attr_init(&attr);
	//pthread_attr_setschedpolicy(&attr, SCHED_RR);
	if (pthread_create(&capture, &attr, (void*)&bttv_capture_thread, &bttv)) {
		ERROR("starting capture thread: %s\n", strerror(errno));
		bttv_uninit(&bttv);
		exit(1);
	}

	if (pthread_create(&lockthread, &attr, (void*)lock_thread, &lock)) {
		ERROR("starting locking thread: %s\n", strerror(errno));
		bttv_uninit(&bttv);
		exit(1);
	}
	
	for (i = 0; i < 384*288*fields; i++) yuv[i] = 0;
	for (i = 0; i < 384*288*fields; i++) 
		yuv[i+384*288*fields] = 128;

	printf("YUV4MPEG2 W384 H%d F25000000:1000000 I%c A4:3\n",
			fields == 2 ? 576 : 288, fields == 2 ? 't' : 'p');

	while (1) {
		/* input */
		in = lock_getnext(&lock, &offset);

		if (debug) {
			if (offset < 544320 && offset > -30) {
				for (i = 0; i < 300; i++) {
					for (j = 0; j < 100; j++) {
						*(out + 2*j*screen->pitch + i) =
							*(in + offset + j*2268 +
									1700
									+ i)*2;
						*(out + (2*j+1)*screen->pitch 
							+ i) = *(in + 1064960 
							+ offset + j*2268 + 
							1700 + i)*2;
					}
				}
				*(out + 2*POLY*screen->pitch + POLX - 1700) = 
					255;
				SDL_UpdateRect(screen, 0, 0, 0, 0);
			}
		}

		for (k = 0; k < fields; k++) {
			field = in + k*1064960;
			poldetect(field, offset);

			/* enable software corrections for small drift */
			if (offset < -20 || offset > 20) offset = 0;
			decodeFrameEsnelColorMMX(yuv + k*384, 
					yuv + 384*288*fields + k*192,  
					yuv + 384*432*fields + k*192,  
					field+4*2268, offset+20, fields*384);
		}

		/* output */
		yuv4mpeg2_out(yuv, fields, stdout);
	}

	pthread_join(capture, NULL);
	pthread_join(lockthread, NULL);
	pthread_attr_destroy(&attr);

	bttv_uninit(&bttv);
	exit(0);
}
	
